import logging
import pathlib
from subprocess import Popen, TimeoutExpired
from typing import Optional, Sequence
from unittest import mock

import pytest

from parsl import HighThroughputExecutor, curvezmq
from parsl.serialize.facade import pack_apply_message, unpack_apply_message

_MOCK_BASE = "parsl.executors.high_throughput.executor"


@pytest.fixture
def encrypted(request: pytest.FixtureRequest):
    if hasattr(request, "param"):
        return request.param
    return True


@pytest.fixture
def htex(encrypted: bool, tmpd_cwd):
    htex = HighThroughputExecutor(encrypted=encrypted)
    htex.max_workers_per_node = 1
    htex.run_dir = tmpd_cwd
    htex.provider.script_dir = tmpd_cwd

    yield htex

    if hasattr(htex, "outgoing_q"):
        htex.scale_in(blocks=1000)
    htex.shutdown()


@pytest.mark.local
@pytest.mark.parametrize("encrypted", (True, False), indirect=True)
@pytest.mark.parametrize("cert_dir_provided", (True, False))
def test_htex_start_encrypted(
    encrypted: bool,
    cert_dir_provided: bool,
    htex: HighThroughputExecutor,
    tmpd_cwd: pathlib.Path,
):
    htex.run_dir = str(tmpd_cwd)
    if cert_dir_provided:
        provided_base_dir = tmpd_cwd / "provided"
        provided_base_dir.mkdir()
        cert_dir = curvezmq.create_certificates(provided_base_dir)
        htex.cert_dir = cert_dir
    else:
        cert_dir = curvezmq.create_certificates(htex.logdir)

    if not encrypted and cert_dir_provided:
        with pytest.raises(AttributeError) as pyt_e:
            htex.start()
        assert "change cert_dir to None" in str(pyt_e.value)
        return

    htex.start()

    assert htex.encrypted is encrypted
    if encrypted:
        assert htex.cert_dir == cert_dir
        assert htex.outgoing_q.zmq_context.cert_dir == cert_dir
        assert htex.incoming_q.zmq_context.cert_dir == cert_dir
        assert htex.command_client.zmq_context.cert_dir == cert_dir
        assert isinstance(htex.outgoing_q.zmq_context, curvezmq.ClientContext)
        assert isinstance(htex.incoming_q.zmq_context, curvezmq.ClientContext)
        assert isinstance(htex.command_client.zmq_context, curvezmq.ClientContext)
    else:
        assert htex.cert_dir is None
        assert htex.outgoing_q.zmq_context.cert_dir is None
        assert htex.incoming_q.zmq_context.cert_dir is None
        assert htex.command_client.zmq_context.cert_dir is None


@pytest.mark.local
@pytest.mark.parametrize("started", (True, False))
@pytest.mark.parametrize("timeout_expires", (True, False))
def test_htex_shutdown(
    started: bool,
    timeout_expires: bool,
    htex: HighThroughputExecutor,
    caplog
):
    mock_ix_proc = mock.Mock(spec=Popen)

    if started:
        htex.interchange_proc = mock_ix_proc

    # This will, in the absence of any exit trigger, block forever if
    # no timeout is given and if the interchange does not terminate.
    # Raise an exception to report that, rather than actually block,
    # and hope that nothing is catching that exception.

    # this function implements the behaviour if the interchange has
    # not received a termination call
    def proc_wait_alive(timeout):
        if timeout:
            raise TimeoutExpired(cmd="mock-interchange", timeout=timeout)
        else:
            raise RuntimeError("This wait call would hang forever")

    def proc_wait_terminated(timeout):
        return 0

    mock_ix_proc.wait.side_effect = proc_wait_alive

    if not timeout_expires:
        # Simulate termination of the Interchange process
        def kill_interchange(*args, **kwargs):
            mock_ix_proc.wait.side_effect = proc_wait_terminated

        mock_ix_proc.terminate.side_effect = kill_interchange

    with caplog.at_level(logging.INFO):
        htex.shutdown()

    if started:
        assert mock_ix_proc.terminate.called
        assert mock_ix_proc.wait.called
        assert {"timeout": 10} == mock_ix_proc.wait.call_args[1]
        if timeout_expires:
            assert "Unable to terminate Interchange" in caplog.text
            assert mock_ix_proc.kill.called
        assert "Attempting HighThroughputExecutor shutdown" in caplog.text
        assert "Finished HighThroughputExecutor shutdown" in caplog.text
    else:
        assert not mock_ix_proc.terminate.called
        assert not mock_ix_proc.wait.called
        assert "HighThroughputExecutor has not started" in caplog.text


@pytest.mark.local
@pytest.mark.parametrize("cmd", (None, "custom-launch-cmd"))
def test_htex_worker_pool_launch_cmd(cmd: Optional[str]):
    if cmd:
        htex = HighThroughputExecutor(launch_cmd=cmd)
        assert htex.launch_cmd == cmd
    else:
        htex = HighThroughputExecutor()
        assert htex.launch_cmd.startswith("process_worker_pool.py")


@pytest.mark.local
@pytest.mark.parametrize("cmd", (None, ["custom", "launch", "cmd"]))
def test_htex_interchange_launch_cmd(cmd: Optional[Sequence[str]]):
    if cmd:
        htex = HighThroughputExecutor(interchange_launch_cmd=cmd)
        assert htex.interchange_launch_cmd == cmd
    else:
        htex = HighThroughputExecutor()
        assert htex.interchange_launch_cmd == ["interchange.py"]


def dyn_exec(buf, *vec_y):
    f, a, _ = unpack_apply_message(buf)
    custom_args = [a, vec_y]
    return f(*custom_args)


@pytest.mark.local
def test_worker_dynamic_import(htex: HighThroughputExecutor):
    def _dot_prod(vec_x, vec_y):
        return sum(x * y for x, y in zip(vec_x, vec_y))

    htex.start()
    htex.scale_out_facade(1)

    num_array = tuple(range(10))

    fn_buf = pack_apply_message(_dot_prod, num_array, {})
    ctxt = {
        "task_executor": {
            "f": f"{dyn_exec.__module__}.{dyn_exec.__name__}",
            "a": num_array,  # prove "custom" dyn_exec
        }
    }
    val = htex.submit_payload(ctxt, fn_buf).result()
    exp_val = _dot_prod(num_array, num_array)

    assert val == exp_val



import pytest

from parsl.executors.high_throughput.manager_record import ManagerRecord
from parsl.executors.high_throughput.manager_selector import BlockIdManagerSelector


@pytest.mark.local
def test_sort_managers():
    ready_managers = {
        b'manager1': {'block_id': 1},
        b'manager2': {'block_id': None},
        b'manager3': {'block_id': 3},
        b'manager4': {'block_id': 2}
    }

    manager_list = {b'manager1', b'manager2', b'manager3', b'manager4'}
    expected_sorted_list = [b'manager2', b'manager1', b'manager4', b'manager3']
    manager_selector = BlockIdManagerSelector()
    sorted_managers = manager_selector.sort_managers(ready_managers, manager_list)
    assert sorted_managers == expected_sorted_list



import random
from abc import ABCMeta, abstractmethod
from typing import Dict, List, Set

from parsl.executors.high_throughput.manager_record import ManagerRecord


class ManagerSelector(metaclass=ABCMeta):

    @abstractmethod
    def sort_managers(self, ready_managers: Dict[bytes, ManagerRecord], manager_list: Set[bytes]) -> List[bytes]:
        """ Sort a given list of managers.

        Any operations pertaining to the sorting and rearrangement of the
        interesting_managers Set should be performed here.
        """
        pass


class RandomManagerSelector(ManagerSelector):

    """Returns a shuffled list of interesting_managers

    By default this strategy is used by the interchange. Works well
    in distributing workloads equally across all availble compute
    resources. The random workload strategy is not effective in
    conjunction with elastic scaling behavior as the even task
    distribution does not allow the scaling down of blocks, leading
    to wasted resource consumption.
    """

    def sort_managers(self, ready_managers: Dict[bytes, ManagerRecord], manager_list: Set[bytes]) -> List[bytes]:
        c_manager_list = list(manager_list)
        random.shuffle(c_manager_list)
        return c_manager_list


class BlockIdManagerSelector(ManagerSelector):

    """Returns an interesting_managers list sorted by block ID

    Observations:
    1. BlockID manager selector helps with workloads that see a varying
    amount of tasks over time. New blocks are prioritized with the
    blockID manager selector, when used with 'htex_auto_scaling', results
    in compute cost savings.

    2. Doesn't really work with bag-of-tasks workloads. When all the tasks
    are put into the queue upfront, all blocks operate at near full
    utilization for the majority of the workload, which task goes where
    doesn't really matter.
    """

    def sort_managers(self, ready_managers: Dict[bytes, ManagerRecord], manager_list: Set[bytes]) -> List[bytes]:
        return sorted(manager_list, key=lambda x: (ready_managers[x]['block_id'] is not None, ready_managers[x]['block_id']))



from parsl.addresses import address_by_interface
from parsl.config import Config
from parsl.dataflow.memoization import BasicMemoizer
from parsl.executors import HighThroughputExecutor
from parsl.launchers import MpiRunLauncher
from parsl.monitoring.monitoring import MonitoringHub
from parsl.providers import PBSProProvider
from parsl.usage_tracking.levels import LEVEL_1

config = Config(
        executors=[
            HighThroughputExecutor(
                label="htex",
                heartbeat_period=15,
                heartbeat_threshold=120,
                worker_debug=True,
                max_workers_per_node=4,
                address=address_by_interface('ib0'),
                provider=PBSProProvider(
                    launcher=MpiRunLauncher(),
                    # PBS directives (header lines): for array jobs pass '-J' option
                    scheduler_options='#PBS -J 1-10',
                    # Command to be run before starting a worker, such as:
                    # 'module load Anaconda; source activate parsl_env'.
                    worker_init='',
                    # number of compute nodes allocated for each block
                    nodes_per_block=3,
                    min_blocks=3,
                    max_blocks=5,
                    cpus_per_node=24,
                    # medium queue has a max walltime of 24 hrs
                    walltime='24:00:00'
                ),
            ),
        ],
        monitoring=MonitoringHub(
            hub_address=address_by_interface('ib0'),
            resource_monitoring_interval=10,
        ),
        strategy='simple',
        retries=3,
        memoizer=BasicMemoizer(checkpoint_mode='task_exit'),
        usage_tracking=LEVEL_1,
)



import pytest

from parsl import join_app, python_app
from parsl.dataflow.errors import JoinError

RESULT_CONSTANT = 3


@python_app
def inner_app():
    return RESULT_CONSTANT


@join_app
def outer_app():
    fut = inner_app()
    return fut


@python_app
def add_one(n):
    return n + 1


@python_app
def combine(*args):
    """Wait for an arbitrary list of futures and return them as a list"""
    return list(args)


@join_app
def outer_make_a_dag_combine(n):
    return combine(*(inner_app() for _ in range(n)))


@join_app
def outer_make_a_dag_multi(n):
    return [inner_app() for _ in range(n)]


def test_result_flow():
    f = outer_app()
    assert f.result() == RESULT_CONSTANT


@join_app
def join_wrong_type_app():
    return 3


def test_wrong_type():
    f = join_wrong_type_app()
    with pytest.raises(TypeError):
        f.result()


def test_dependency_on_joined():
    g = add_one(outer_app())
    assert g.result() == RESULT_CONSTANT + 1


def test_combine():
    f = outer_make_a_dag_combine(inner_app())
    assert f.result() == [RESULT_CONSTANT] * RESULT_CONSTANT


def test_multiple_return():
    f = outer_make_a_dag_multi(inner_app())
    assert f.result() == [RESULT_CONSTANT] * RESULT_CONSTANT


class InnerError(RuntimeError):
    pass


@python_app
def inner_error(s="X"):
    raise InnerError("Error " + s)


@join_app
def outer_error():
    return inner_error()


@join_app
def outer_two_errors():
    return [inner_error("A"), inner_error("B")]


@join_app
def outer_one_error_one_result():
    return [inner_error("A"), inner_app()]


def test_error():
    f = outer_error()
    e = f.exception()
    assert isinstance(e, JoinError)

    assert len(e.dependent_exceptions_tids) == 1
    assert isinstance(e.dependent_exceptions_tids[0][0], InnerError)
    assert e.dependent_exceptions_tids[0][1].startswith("task ")


def test_two_errors():
    f = outer_two_errors()
    e = f.exception()
    assert isinstance(e, JoinError)
    assert len(e.dependent_exceptions_tids) == 2

    de0 = e.dependent_exceptions_tids[0][0]
    assert isinstance(de0, InnerError)
    assert de0.args[0] == "Error A"
    assert e.dependent_exceptions_tids[0][1].startswith("task ")

    de1 = e.dependent_exceptions_tids[1][0]
    assert isinstance(de1, InnerError)
    assert de1.args[0] == "Error B"
    assert e.dependent_exceptions_tids[1][1].startswith("task ")


def test_one_error_one_result():
    f = outer_one_error_one_result()
    e = f.exception()

    assert isinstance(e, JoinError)
    assert len(e.dependent_exceptions_tids) == 1

    de0 = e.dependent_exceptions_tids[0][0]
    assert isinstance(de0, InnerError)
    assert de0.args[0] == "Error A"
    assert e.dependent_exceptions_tids[0][1].startswith("task ")


@join_app
def app_no_futures():
    return []


def test_no_futures():
    # tests that a list of futures that contains no futures will
    # complete - regression test for issue #2792
    f = app_no_futures()
    assert f.result() == []



"""Functions used to explain kwargs"""
from pathlib import Path

import pytest

from parsl import File, python_app


def test_inputs():
    @python_app()
    def map_app(x):
        return x * 2

    @python_app()
    def reduce_app(inputs=()):
        return sum(inputs)

    map_futures = [map_app(x) for x in range(3)]
    reduce_future = reduce_app(inputs=map_futures)

    assert reduce_future.result() == 6


@pytest.mark.shared_fs
def test_outputs(tmpd_cwd):
    @python_app()
    def write_app(message, outputs=()):
        """Write a single message to every file in outputs"""
        for path in outputs:
            with open(path, 'w') as fp:
                print(message, file=fp)

    to_write = [
        File(Path(tmpd_cwd) / 'output-0.txt'),
        File(Path(tmpd_cwd) / 'output-1.txt')
    ]
    write_app('Hello!', outputs=to_write).result()
    for path in to_write:
        with open(path) as fp:
            assert fp.read() == 'Hello!\n'



import os

import pytest

from parsl.executors.execute_task import execute_task
from parsl.serialize.facade import pack_apply_message


def addemup(*args: int, name: str = "apples"):
    total = sum(args)
    return f"{total} {name}"


@pytest.mark.local
def test_execute_task():
    args = (1, 2, 3)
    kwargs = {"name": "boots"}
    buff = pack_apply_message(addemup, args, kwargs)
    res = execute_task(buff)
    assert res == addemup(*args, **kwargs)



import logging
import os

from parsl.executors.taskvine.errors import TaskVineFactoryFailure
from parsl.process_loggers import wrap_with_logs

# This try except clause prevents import errors
# when TaskVine is not used in Parsl.
try:
    from ndcctools.taskvine import Factory
    taskvine_available = True
except ImportError:
    taskvine_available = False

logger = logging.getLogger(__name__)


@wrap_with_logs
def _taskvine_factory(should_stop, factory_config):
    if not taskvine_available:
        logger.debug("TaskVine package cannot be found. Please install the ndcctools package.")
        return
    logger.debug("Starting TaskVine factory process")

    try:
        # create the factory according to the project name if given
        if factory_config._project_name:
            factory = Factory(batch_type=factory_config.batch_type,
                              manager_name=factory_config._project_name,
                              )
        else:
            factory = Factory(batch_type=factory_config.batch_type,
                              manager_host_port=f"{factory_config._project_address}:{factory_config._project_port}",
                              )
    except Exception as e:
        raise TaskVineFactoryFailure(f'Cannot create factory with exception {e}')

    # Set attributes of this factory
    if factory_config._project_password_file:
        factory.password = factory_config._project_password_file
    factory.factory_timeout = factory_config.factory_timeout
    factory.scratch_dir = factory_config.scratch_dir
    factory.min_workers = factory_config.min_workers
    factory.max_workers = factory_config.max_workers
    factory.workers_per_cycle = factory_config.workers_per_cycle

    # create scratch dir if factory process gets ahead of the manager.
    os.makedirs(factory.scratch_dir, exist_ok=True)

    if factory_config.worker_options:
        factory.extra_options = factory_config.worker_options
    factory.timeout = factory_config.worker_timeout
    if factory_config.cores:
        factory.cores = factory_config.cores
    if factory_config.gpus:
        factory.gpus = factory_config.gpus
    if factory_config.memory:
        factory.memory = factory_config.memory
    if factory_config.disk:
        factory.disk = factory_config.disk
    if factory_config.python_env:
        factory.python_env = factory_config.python_env

    if factory_config.condor_requirements:
        factory.condor_requirements = factory_config.condor_requirements
    if factory_config.batch_options:
        factory.batch_options = factory_config.batch_options

    # run factory through Python context and wait for signal to stop.
    with factory:
        should_stop.wait()

    logger.debug("Exiting TaskVine factory process")



import inspect
from abc import abstractmethod
from functools import singledispatch
from typing import Any, List, Sequence

from parsl.utils import RepresentationMixin

# Traverse the configuration hierarchy, returning a JSON component
# for each one. Configuration components which implement
# RepresentationMixin will be in the right form for inspecting
# object attributes. Configuration components which are lists or tuples
# are traversed in sequence. Other types default to reporting no
# usage information.


@singledispatch
def get_parsl_usage(obj) -> List[Any]:
    return []


@get_parsl_usage.register
def get_parsl_usage_representation_mixin(obj: RepresentationMixin) -> List[Any]:
    t = type(obj)
    qualified_name = t.__module__ + "." + t.__name__

    # me can contain anything that can be rendered as JSON
    me: List[Any] = []

    if isinstance(obj, UsageInformation):
        # report rich usage information for this component
        attrs = {'c': qualified_name}
        attrs.update(obj.get_usage_information())
        me = [attrs]
    else:
        # report the class name of this component
        me = [qualified_name]

    # unwrap typeguard-style unwrapping
    init: Any = type(obj).__init__
    if hasattr(init, '__wrapped__'):
        init = init.__wrapped__

    argspec = inspect.getfullargspec(init)

    for arg in argspec.args[1:]:  # skip first arg, self
        arg_value = getattr(obj, arg)
        d = get_parsl_usage(arg_value)
        me += d

    return me


@get_parsl_usage.register(list)
@get_parsl_usage.register(tuple)
def get_parsl_usage_sequence(obj: Sequence) -> List[Any]:
    result = []
    for v in obj:
        result += get_parsl_usage(v)
    return result


class UsageInformation:
    @abstractmethod
    def get_usage_information(self) -> dict:
        pass



import argparse

import pytest

import parsl
from parsl.app.app import python_app
from parsl.config import Config
from parsl.dataflow.memoization import BasicMemoizer
from parsl.dataflow.taskrecord import TaskRecord


class DontReuseSevenMemoizer(BasicMemoizer):
    def check_memo(self, task_record: TaskRecord):
        if task_record['args'][0] == 7:
            return None  # we didn't find a suitable memo record...
        else:
            return super().check_memo(task_record)


def local_config():
    return Config(memoizer=DontReuseSevenMemoizer())


@python_app(cache=True)
def random_uuid(x, cache=True):
    import uuid
    return str(uuid.uuid4())


@pytest.mark.local
@pytest.mark.parametrize("n", (10,))
def test_python_memoization(n: int):
    """Testing using an alternate memoization plugin.

    The plugin under test memoizes except for functions with the parameter 7,
    and that behaviour should be observable in uuid app results.
    """

    x = random_uuid(0).result()

    for i in range(0, n):
        foo = random_uuid(0)
        assert foo.result() == x, "Memoized results were incorrectly not used"

    y = random_uuid(7).result()

    for i in range(0, n):
        foo = random_uuid(7)
        assert foo.result() != y, "Memoized results were incorrectly used"



"""
The aim of this configuration is to run a local htex
in a similar manner to htex_local.py, but with lots of
options different and more complicated than in that
configuration, so that more code paths are executed
than when testing only with htex_local.

It does not matter too much *what* is different in this
configuration; what matters is that the differences
cause significantly different pieces of parsl code to be
run - for example, by turning on monitoring, by allowing
blocks to be started by a strategy, by using a different
set of staging providers, by using timing parameters that
will cause substantially different behaviour on whatever
those timing parameters control.
"""

import os

from parsl.config import Config
from parsl.data_provider.file_noop import NoOpFileStaging
from parsl.data_provider.ftp import FTPInTaskStaging
from parsl.data_provider.http import HTTPInTaskStaging
from parsl.data_provider.zip import ZipFileStaging
from parsl.dataflow.memoization import BasicMemoizer
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SingleNodeLauncher
from parsl.monitoring import MonitoringHub
from parsl.providers import LocalProvider

working_dir = os.getcwd() + "/" + "test_htex_alternate"


def fresh_config():
    return Config(
        executors=[
            HighThroughputExecutor(
                address="127.0.0.1",
                label="htex_Local",
                working_dir=working_dir,
                storage_access=[ZipFileStaging(), FTPInTaskStaging(), HTTPInTaskStaging(), NoOpFileStaging()],
                worker_debug=True,
                cores_per_worker=1,
                heartbeat_period=2,
                heartbeat_threshold=5,
                poll_period=100,
                encrypted=True,
                provider=LocalProvider(
                    init_blocks=0,
                    min_blocks=0,
                    max_blocks=5,
                    launcher=SingleNodeLauncher(),
                ),
                block_error_handler=False
            )
        ],
        strategy='simple',
        memoizer=BasicMemoizer(memoize=True, checkpoint_mode='task_exit'),
        retries=2,
        monitoring=MonitoringHub(
                        monitoring_debug=False,
                        resource_monitoring_interval=1,
        ),
        usage_tracking=3,
        project_name="parsl htex_local_alternate test configuration"
    )


config = fresh_config()



template_string = '''#!/bin/bash

#BSUB -W ${walltime}
#BSUB -J ${jobname}
#BSUB -o ${submit_script_dir}/${jobname}.submit.stdout
#BSUB -e ${submit_script_dir}/${jobname}.submit.stderr
${scheduler_options}

${worker_init}

export JOBNAME="${jobname}"

$user_script
'''



def read_sort_write(in_path, out_path):
    with open(in_path) as u:
        strs = u.read().split()
        strs.sort()
    with open(out_path, 'w') as s:
        for e in strs:
            print(e, file=s)



import json
import sys
from typing import List

_setup_paths: List[str] = []
try:
    import radical.pilot as rp
    import radical.utils as ru
except ImportError:
    pass


MPI = "mpi"
RP_ENV = "rp"
CLIENT = "client"
RPEX_ENV = "ve_rpex"
MPI_WORKER = "MPIWorker"
DEFAULT_WORKER = "DefaultWorker"


class ResourceConfig:
    """
    This ResourceConfig class is an abstraction of the resource
    configuration of the RAPTOR layer in the RADICAL-Pilot runtime system.

    This class sets up the default configuration values for the executor and
    allows the user to specify different resource requirements flexibly.

    For more information:
    https://radicalpilot.readthedocs.io/en/stable/tutorials/raptor.html

    Parameters
    ----------
    masters : int
        The number of masters to be deployed by RAPTOR.
        Default is 1.

    workers : int
        The number of workers to be deployed by RAPTOR.
        Default is 1.

    worker_gpus_per_node : int
        The number of GPUs a worker will operate on per node.
        Default is 0.

    worker_cores_per_node : int
        The number of CPU cores a worker will operate on per node.
        Default is 4.

    cores_per_master : int
        The number of cores a master will operate on per node.
        Default is 1.

    nodes_per_worker : int
        The number of nodes to be occupied by every worker.
        Default is 1.

    pilot_env_path : str
        The path to an exisitng pilot environment.
        Default is an empty string (RADICAL-Pilot will create one).

    pilot_env_name : str
        The name of the pilot environment.
        Default is "ve_rpex".

    pilot_env_pre_exec : list
        List of commands to be executed before starting the pilot environment.
        Default is an empty list.

    pilot_env_type : str
        The type of the pilot environment (e.g., 'venv', 'conda').
        Default is "venv".

    pilot_env_setup : list
        List of setup commands/packages for the pilot environment.
        Default is an empty list.

    python_v : str
        The Python version to be used in the pilot environment.
        Default is determined by the system's Python version.

    worker_type : str
        The type of worker(s) to be deployed by RAPTOR on the compute
        resources.
        Default is "DefaultWorker".
    """

    masters: int = 1
    workers: int = 1

    worker_gpus_per_node: int = 0
    worker_cores_per_node: int = 4

    cores_per_master: int = 1
    nodes_per_worker: int = 1

    pilot_env_mode: str = CLIENT
    pilot_env_path: str = ""
    pilot_env_type: str = "venv"
    pilot_env_name: str = RP_ENV
    pilot_env_pre_exec: List[str] = []
    pilot_env_setup: List[str] = _setup_paths

    python_v: str = f'{sys.version_info[0]}.{sys.version_info[1]}'
    worker_type: str = DEFAULT_WORKER

    def get_config(cls, path=None):

        # Default ENV mode for RP is to reuse
        # the client side. If this is not the case,
        # then RP will create a new env named ve_rpex
        # The user need to make sure that under:
        # $HOME/.radical/pilot/configs/*_resource.json
        # that virtenv_mode = local
        if cls.pilot_env_mode != CLIENT:
            cls.pilot_env_name = RPEX_ENV

        if MPI in cls.worker_type.lower() and \
           "mpi4py" not in cls.pilot_env_setup:
            cls.pilot_env_setup.append("mpi4py")

        cfg = {
            'n_masters': cls.masters,
            'n_workers': cls.workers,
            'worker_type': cls.worker_type,
            'gpus_per_node': cls.worker_gpus_per_node,
            'cores_per_node': cls.worker_cores_per_node,
            'cores_per_master': cls.cores_per_master,
            'nodes_per_worker': cls.nodes_per_worker,

            'pilot_env': {
                "version": cls.python_v,
                "name": cls.pilot_env_name,
                "path": cls.pilot_env_path,
                "type": cls.pilot_env_type,
                "setup": cls.pilot_env_setup,
                "pre_exec": cls.pilot_env_pre_exec
            },

            'pilot_env_mode': cls.pilot_env_mode,

            'master_descr': {
                "ranks": 1,
                "cores_per_rank": 1,
                "mode": rp.RAPTOR_MASTER,
                "named_env": cls.pilot_env_name,
            },

            'worker_descr': {
                "mode": rp.RAPTOR_WORKER,
                "named_env": cls.pilot_env_name,
                "raptor_file": "./rpex_worker.py",
                "raptor_class": cls.worker_type if
                cls.worker_type.lower() != MPI else MPI_WORKER,
                "ranks": cls.nodes_per_worker * cls.worker_cores_per_node,
                "gpus_per_rank": cls.nodes_per_worker * cls.worker_gpus_per_node,
            }}

        # Convert the class instance to a Json file or a Config dict.
        if path:
            config_path = 'rpex.cfg'
            config_path = path + '/' + config_path
            with open(config_path, 'w') as f:
                json.dump(cfg, f, indent=4)
        else:
            config_obj = ru.Config(from_dict=cfg)
            return config_obj



import logging
import math
import os
import re
import sys
import time
from typing import Any, Dict, Optional

import typeguard

from parsl.jobs.states import JobState, JobStatus
from parsl.launchers import SingleNodeLauncher
from parsl.launchers.base import Launcher
from parsl.providers.cluster_provider import ClusterProvider
from parsl.providers.errors import SubmitException
from parsl.providers.slurm.template import template_string
from parsl.utils import RepresentationMixin, wtime_to_minutes

logger = logging.getLogger(__name__)

# From https://slurm.schedmd.com/sacct.html#SECTION_JOB-STATE-CODES
sacct_translate_table = {
    'PENDING': JobState.PENDING,
    'RUNNING': JobState.RUNNING,
    'CANCELLED': JobState.CANCELLED,
    'COMPLETED': JobState.COMPLETED,
    'FAILED': JobState.FAILED,
    'NODE_FAIL': JobState.FAILED,
    'BOOT_FAIL': JobState.FAILED,
    'DEADLINE': JobState.TIMEOUT,
    'TIMEOUT': JobState.TIMEOUT,
    'REVOKED': JobState.FAILED,
    'OUT_OF_MEMORY': JobState.FAILED,
    'SUSPENDED': JobState.HELD,
    'PREEMPTED': JobState.TIMEOUT,
    'REQUEUED': JobState.PENDING
}

squeue_translate_table = {
    'PD': JobState.PENDING,
    'R': JobState.RUNNING,
    'CA': JobState.CANCELLED,
    'CF': JobState.PENDING,  # (configuring),
    'CG': JobState.RUNNING,  # (completing),
    'CD': JobState.COMPLETED,
    'F': JobState.FAILED,  # (failed),
    'TO': JobState.TIMEOUT,  # (timeout),
    'NF': JobState.FAILED,  # (node failure),
    'RV': JobState.FAILED,  # (revoked) and
    'SE': JobState.FAILED   # (special exit state)
}


if sys.version_info < (3, 12):
    from itertools import islice
    from typing import Iterable

    def batched(
        iterable: Iterable[tuple[object, Any]], n: int, *, strict: bool = False
    ):
        """Batched
        Turns a list into a batch of size n. This code is from
        https://docs.python.org/3.12/library/itertools.html#itertools.batched
        and in versions 3.12+ this can be replaced with
        itertools.batched
        """
        if n < 1:
            raise ValueError("n must be at least one")
        iterator = iter(iterable)
        while batch := tuple(islice(iterator, n)):
            yield batch
else:
    from itertools import batched


class SlurmProvider(ClusterProvider, RepresentationMixin):
    """Slurm Execution Provider

    This provider uses sbatch to submit, sacct for status and scancel to cancel
    jobs. The sbatch script to be used is created from a template file in this
    same module.

    Parameters
    ----------
    partition : str
        Slurm partition to request blocks from. If unspecified or ``None``, no partition slurm directive will be specified.
    account : str
        Slurm account to which to charge resources used by the job. If unspecified or ``None``, the job will use the
        user's default account.
    qos : str
        Slurm queue to place job in. If unspecified or ``None``, no queue slurm directive will be specified.
    constraint : str
        Slurm job constraint, often used to choose cpu or gpu type. If unspecified or ``None``, no constraint slurm directive will be added.
    clusters : str
        Slurm cluster name, or comma seperated cluster list, used to choose between different clusters in a federated Slurm instance.
        If unspecified or ``None``, no slurm directive for clusters will be added.
    nodes_per_block : int
        Nodes to provision per block.
    cores_per_node : int
        Specify the number of cores to provision per node. If set to None, executors
        will assume all cores on the node are available for computation. Default is None.
    mem_per_node : int
        Specify the real memory to provision per node in GB. If set to None, no
        explicit request to the scheduler will be made. Default is None.
    init_blocks : int
        Number of blocks to provision at the start of the run. Default is 1.
    min_blocks : int
        Minimum number of blocks to maintain.
    max_blocks : int
        Maximum number of blocks to maintain.
    parallelism : float
        Ratio of provisioned task slots to active tasks. A parallelism value of 1 represents aggressive
        scaling where as many resources as possible are used; parallelism close to 0 represents
        the opposite situation in which as few resources as possible (i.e., min_blocks) are used.
    walltime : str
        Walltime requested per block in HH:MM:SS.
    scheduler_options : str
        String to prepend to the #SBATCH blocks in the submit script to the scheduler.
    regex_job_id : str
        The regular expression used to extract the job ID from the ``sbatch`` standard output.
        The default is ``r"Submitted batch job (?P<id>\\S*)"``, where ``id`` is the regular expression
        symbolic group for the job ID.
    worker_init : str
        Command to be run before starting a worker, such as 'module load Anaconda; source activate env'.
    cmd_timeout : int (Default = 10)
        Number of seconds to wait for slurm commands to finish. For schedulers with many this
        may need to be increased to wait longer for scheduler information.
    status_batch_size: int (Default = 50)
        Number of jobs to batch together in calls to the scheduler status. For schedulers
        with many jobs this may need to be decreased to get jobs in smaller batches.
    exclusive : bool (Default = True)
        Requests nodes which are not shared with other running jobs.
    launcher : Launcher
        Launcher for this provider. Possible launchers include
        :class:`~parsl.launchers.SingleNodeLauncher` (the default),
        :class:`~parsl.launchers.SrunLauncher`, or
        :class:`~parsl.launchers.AprunLauncher`
    """

    @typeguard.typechecked
    def __init__(self,
                 partition: Optional[str] = None,
                 account: Optional[str] = None,
                 qos: Optional[str] = None,
                 constraint: Optional[str] = None,
                 clusters: Optional[str] = None,
                 nodes_per_block: int = 1,
                 cores_per_node: Optional[int] = None,
                 mem_per_node: Optional[int] = None,
                 init_blocks: int = 1,
                 min_blocks: int = 0,
                 max_blocks: int = 1,
                 parallelism: float = 1,
                 walltime: str = "00:10:00",
                 scheduler_options: str = '',
                 regex_job_id: str = r"Submitted batch job (?P<id>\S*)",
                 worker_init: str = '',
                 cmd_timeout: int = 10,
                 status_batch_size: int = 50,
                 exclusive: bool = True,
                 launcher: Launcher = SingleNodeLauncher()):
        label = 'slurm'
        super().__init__(label,
                         nodes_per_block,
                         init_blocks,
                         min_blocks,
                         max_blocks,
                         parallelism,
                         walltime,
                         cmd_timeout=cmd_timeout,
                         launcher=launcher)

        self.partition = partition
        self.cores_per_node = cores_per_node
        self.mem_per_node = mem_per_node
        self.exclusive = exclusive
        self.account = account
        self.qos = qos
        self.constraint = constraint
        self.clusters = clusters
        # Used to batch requests to sacct/squeue for long jobs lists
        self.status_batch_size = status_batch_size
        self.scheduler_options = scheduler_options + '\n'
        if exclusive:
            self.scheduler_options += "#SBATCH --exclusive\n"
        if partition:
            self.scheduler_options += "#SBATCH --partition={}\n".format(partition)
        if account:
            self.scheduler_options += "#SBATCH --account={}\n".format(account)
        if qos:
            self.scheduler_options += "#SBATCH --qos={}\n".format(qos)
        if constraint:
            self.scheduler_options += "#SBATCH --constraint={}\n".format(constraint)
        if clusters:
            self.scheduler_options += "#SBATCH --clusters={}\n".format(clusters)

        self.regex_job_id = regex_job_id
        self.worker_init = worker_init + '\n'
        # Check if sacct works and if not fall back to squeue
        cmd = "sacct -X"
        logger.debug("Executing %s", cmd)
        retcode, stdout, stderr = self.execute_wait(cmd)
        # If sacct fails it should return retcode=1 stderr="Slurm accounting storage is disabled"
        logger.debug(f"sacct returned retcode={retcode} stderr={stderr}")
        if retcode == 0:
            logger.debug("using sacct to get job status")
            _cmd = "sacct"
            # Add clusters option to sacct if provided
            if self.clusters:
                _cmd += f" --clusters={self.clusters}"
            # Using state%20 to get enough characters to not truncate output
            # of the state. Without output can look like "<job_id>     CANCELLED+"
            self._cmd = _cmd + " -X --noheader --format=jobid,state%20 --job '{0}'"
            self._translate_table = sacct_translate_table
        else:
            logger.debug(f"sacct failed with retcode={retcode}")
            logger.debug("falling back to using squeue to get job status")
            _cmd = "squeue"
            # Add clusters option to squeue if provided
            if self.clusters:
                _cmd += f" --clusters={self.clusters}"
            self._cmd = _cmd + " --noheader --format='%i %t' --job '{0}'"
            self._translate_table = squeue_translate_table

    def _status(self):
        '''Returns the status list for a list of job_ids

        Args:
              self

        Returns:
              [status...] : Status list of all jobs
        '''
        active_jobs = {jid: job for jid, job in self.resources.items() if not job["status"].terminal}
        if len(active_jobs) == 0:
            logger.debug("No active jobs, skipping status update")
            return

        stdout = ""
        for job_batch in batched(active_jobs.items(), self.status_batch_size):
            job_id_list = ",".join(jid for jid, job in job_batch if not job["status"].terminal)
            cmd = self._cmd.format(job_id_list)
            logger.debug("Executing %s", cmd)
            retcode, batch_stdout, batch_stderr = self.execute_wait(cmd)
            logger.debug(f"sacct/squeue returned {retcode} {batch_stdout} {batch_stderr}")
            stdout += batch_stdout
            # Execute_wait failed. Do no update
            if retcode != 0:
                logger.warning("sacct/squeue failed with non-zero exit code {}".format(retcode))
                return

        jobs_missing = set(self.resources.keys())
        for line in stdout.split('\n'):
            if not line:
                # Blank line
                continue
            # Sacct includes extra information in some outputs
            # For example "<job_id> CANCELLED by <user_id>"
            # This splits and ignores anything past the first two unpacked values
            job_id, slurm_state, *ignore = line.split()
            if slurm_state not in self._translate_table:
                logger.warning(f"Slurm status {slurm_state} is not recognized")
            status = self._translate_table.get(slurm_state, JobState.UNKNOWN)
            logger.debug("Updating job {} with slurm status {} to parsl state {!s}".format(job_id, slurm_state, status))
            self.resources[job_id]['status'] = JobStatus(status,
                                                         stdout_path=self.resources[job_id]['job_stdout_path'],
                                                         stderr_path=self.resources[job_id]['job_stderr_path'])
            jobs_missing.remove(job_id)

        # sacct can get job info after jobs have completed so this path shouldn't be hit
        # squeue does not report on jobs that are not running. So we are filling in the
        # blanks for missing jobs, we might lose some information about why the jobs failed.
        for missing_job in jobs_missing:
            logger.debug("Updating missing job {} to completed status".format(missing_job))
            self.resources[missing_job]['status'] = JobStatus(
                JobState.COMPLETED, stdout_path=self.resources[missing_job]['job_stdout_path'],
                stderr_path=self.resources[missing_job]['job_stderr_path'])

    def submit(self, command: str, tasks_per_node: int, job_name="parsl.slurm") -> str:
        """Submit the command as a slurm job.

        Parameters
        ----------
        command : str
            Command to be made on the remote side.
        tasks_per_node : int
            Command invocations to be launched per node
        job_name : str
            Name for the job
        Returns
        -------
        job id : str
            A string identifier for the job
        """

        scheduler_options = self.scheduler_options
        worker_init = self.worker_init
        if self.mem_per_node is not None:
            scheduler_options += '#SBATCH --mem={}g\n'.format(self.mem_per_node)
            worker_init += 'export PARSL_MEMORY_GB={}\n'.format(self.mem_per_node)
        if self.cores_per_node is not None:
            cpus_per_task = math.floor(self.cores_per_node / tasks_per_node)
            scheduler_options += '#SBATCH --cpus-per-task={}'.format(cpus_per_task)
            worker_init += 'export PARSL_CORES={}\n'.format(cpus_per_task)

        job_name = "{0}.{1}".format(job_name, time.time())

        assert self.script_dir, "Expected script_dir to be set"
        script_path = os.path.join(self.script_dir, job_name)
        script_path = os.path.abspath(script_path)
        job_stdout_path = script_path + ".stdout"
        job_stderr_path = script_path + ".stderr"

        logger.debug("Requesting one block with {} nodes".format(self.nodes_per_block))

        job_config: Dict[str, Any] = {}
        job_config["submit_script_dir"] = self.script_dir
        job_config["nodes"] = self.nodes_per_block
        job_config["tasks_per_node"] = tasks_per_node
        job_config["walltime"] = wtime_to_minutes(self.walltime)
        job_config["scheduler_options"] = scheduler_options
        job_config["worker_init"] = worker_init
        job_config["user_script"] = command
        job_config["job_stdout_path"] = job_stdout_path
        job_config["job_stderr_path"] = job_stderr_path

        # Wrap the command
        job_config["user_script"] = self.launcher(command,
                                                  tasks_per_node,
                                                  self.nodes_per_block)

        logger.debug("Writing submit script")
        self._write_submit_script(template_string, script_path, job_name, job_config)

        retcode, stdout, stderr = self.execute_wait("sbatch {0}".format(script_path))

        if retcode == 0:
            for line in stdout.split('\n'):
                match = re.match(self.regex_job_id, line)
                if match:
                    job_id = match.group("id")
                    self.resources[job_id] = {'job_id': job_id,
                                              'status': JobStatus(JobState.PENDING),
                                              'job_stdout_path': job_stdout_path,
                                              'job_stderr_path': job_stderr_path,
                                              }
                    return job_id
            else:
                logger.error("Could not read job ID from submit command standard output.")
                logger.error("Retcode:%s STDOUT:%s STDERR:%s", retcode, stdout.strip(), stderr.strip())
                raise SubmitException(
                    job_name,
                    "Could not read job ID from submit command standard output",
                    stdout=stdout,
                    stderr=stderr,
                    retcode=retcode
                )
        else:
            logger.error("Submit command failed")
            logger.error("Retcode:%s STDOUT:%s STDERR:%s", retcode, stdout.strip(), stderr.strip())
            raise SubmitException(
                job_name, "Could not read job ID from submit command standard output",
                stdout=stdout,
                stderr=stderr,
                retcode=retcode
            )

    def cancel(self, job_ids):
        ''' Cancels the jobs specified by a list of job ids

        Args:
        job_ids : [<job_id> ...]

        Returns :
        [True/False...] : If the cancel operation fails the entire list will be False.
        '''

        job_id_list = ' '.join(job_ids)

        # Make the command to cancel jobs
        _cmd = "scancel"
        if self.clusters:
            _cmd += f" --clusters={self.clusters}"
        _cmd += " {0}"

        retcode, stdout, stderr = self.execute_wait(_cmd.format(job_id_list))
        rets = None
        if retcode == 0:
            for jid in job_ids:
                self.resources[jid]['status'] = JobStatus(JobState.CANCELLED)  # Setting state to cancelled
            rets = [True for i in job_ids]
        else:
            rets = [False for i in job_ids]

        return rets

    @property
    def status_polling_interval(self):
        return 60



from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.providers import LocalProvider
from parsl.usage_tracking.levels import LEVEL_1

config = Config(
    executors=[
        HighThroughputExecutor(
            label="htex_local",
            cores_per_worker=1,
            provider=LocalProvider(
                init_blocks=1,
                max_blocks=1,
            ),
        )
    ],
    usage_tracking=LEVEL_1,
)



from typing import List


class ParslError(Exception):
    """Base class for all exceptions.

    Only to be invoked when a more specific error is not available.
    """


class ConfigurationError(ParslError):
    """Raised when a component receives an invalid configuration.
    """


class OptionalModuleMissing(ParslError):
    ''' Error raised when a required module is missing for a optional/extra component
    '''

    def __init__(self, module_names: List[str], reason: str):
        self.module_names = module_names
        self.reason = reason

    def __str__(self) -> str:
        return "The functionality requested requires optional modules {0} which could not be imported, because: {1}".format(
            self.module_names, self.reason
        )


class InternalConsistencyError(ParslError):
    """Raised when a component enounters an internal inconsistency.
    """


class NoDataFlowKernelError(ParslError):
    """Raised when no DataFlowKernel is available for an operation that needs one.
    """



import logging

import pytest

from parsl.executors.high_throughput.mpi_prefix_composer import (
    compose_all,
    compose_aprun_launch_cmd,
    compose_mpiexec_launch_cmd,
    compose_srun_launch_cmd,
)
from parsl.executors.high_throughput.mpi_resource_management import Scheduler

resource_spec = {"num_nodes": 2,
                 "num_ranks": 8,
                 "ranks_per_node": 4}


@pytest.mark.local
def test_srun_launch_cmd():
    prefix_name, composed_prefix = compose_srun_launch_cmd(
        resource_spec=resource_spec, node_hostnames=["node1", "node2"]
    )
    assert prefix_name == "PARSL_SRUN_PREFIX"
    logging.warning(composed_prefix)

    assert "None" not in composed_prefix


@pytest.mark.local
def test_aprun_launch_cmd():
    prefix_name, composed_prefix = compose_aprun_launch_cmd(
        resource_spec=resource_spec, node_hostnames=["node1", "node2"]
    )
    logging.warning(composed_prefix)
    assert prefix_name == "PARSL_APRUN_PREFIX"
    assert "None" not in composed_prefix


@pytest.mark.local
def test_mpiexec_launch_cmd():
    prefix_name, composed_prefix = compose_mpiexec_launch_cmd(
        resource_spec=resource_spec, node_hostnames=["node1", "node2"]
    )
    logging.warning(composed_prefix)
    assert prefix_name == "PARSL_MPIEXEC_PREFIX"
    assert "None" not in composed_prefix


@pytest.mark.local
def test_slurm_launch_cmd():
    table = compose_all(
        mpi_launcher="srun",
        resource_spec=resource_spec,
        node_hostnames=["NODE001", "NODE002"],
    )

    assert "PARSL_MPI_PREFIX" in table
    assert "PARSL_SRUN_PREFIX" in table


@pytest.mark.local
def test_default_launch_cmd():
    table = compose_all(
        mpi_launcher="srun",
        resource_spec=resource_spec,
        node_hostnames=["NODE001", "NODE002"],
    )

    assert "PARSL_MPI_PREFIX" in table
    assert "PARSL_SRUN_PREFIX" in table
    assert table["PARSL_MPI_PREFIX"] == table["PARSL_SRUN_PREFIX"]



import functools
import logging
import pickle

import dill

logger = logging.getLogger(__name__)
from typing import Any

from parsl.serialize.base import SerializerBase


class PickleSerializer(SerializerBase):
    """ Pickle serialization covers most python objects, with some notable exceptions:

    * functions defined in a interpreter/notebook
    * classes defined in local context and not importable using a fully qualified name
    * closures, generators and coroutines
    """

    identifier = b'01'

    def serialize(self, data: Any) -> bytes:
        return pickle.dumps(data)

    def deserialize(self, body: bytes) -> Any:
        return pickle.loads(body)


class DillSerializer(SerializerBase):
    """ Dill serialization works on a superset of object including the ones covered by pickle.
    However for most cases pickle is faster. For most callable objects the additional overhead
    of dill can be amortized with an lru_cache. Here's items that dill handles that pickle
    doesn't:

    * functions defined in a interpreter/notebook
    * classes defined in local context and not importable using a fully qualified name
    * functions that are wrapped/decorated by other functions/classes
    * closures
    """

    identifier = b'02'

    def serialize(self, data: Any) -> bytes:
        return dill.dumps(data)

    def deserialize(self, body: bytes) -> Any:
        return dill.loads(body)


class DillCallableSerializer(SerializerBase):
    """This serializer is a variant of the DillSerializer that will
    serialize and deserialize callables using an lru_cache, under the
    assumption that callables are immutable and so can be cached.
    """

    identifier = b'C2'

    @functools.lru_cache
    def serialize(self, data: Any) -> bytes:
        return dill.dumps(data)

    @functools.lru_cache
    def deserialize(self, body: bytes) -> Any:
        return dill.loads(body)



import pytest

from parsl import AUTO_LOGNAME, Config, bash_app, python_app
from parsl.executors import ThreadPoolExecutor


def local_config():
    return Config(executors=[ThreadPoolExecutor()])


@pytest.mark.local
def test_default_inputs():
    @python_app
    def identity(inp):
        return inp

    @bash_app
    def sum_inputs(inputs=[identity(1), identity(2)], stdout=AUTO_LOGNAME):
        calc = sum(inputs)
        return f"echo {calc}"

    fut = sum_inputs()
    fut.result()
    with open(fut.stdout, 'r') as f:
        assert int(f.read()) == 3



import datetime
import logging
import os
import time
from functools import wraps
from multiprocessing import Event
from typing import Any, Callable, Dict, List, Sequence, Tuple

from parsl.monitoring.message_type import MessageType
from parsl.monitoring.radios.base import RadioConfig
from parsl.multiprocessing import ForkProcess
from parsl.process_loggers import wrap_with_logs

logger = logging.getLogger(__name__)


def monitor_wrapper(*,
                    f: Any,           # per app
                    args: Sequence,   # per invocation
                    kwargs: Dict,     # per invocation
                    x_try_id: int,    # per invocation
                    x_task_id: int,   # per invocation
                    radio_config: RadioConfig,   # per executor
                    run_id: str,      # per workflow
                    logging_level: int,  # per workflow
                    sleep_dur: float,  # per workflow
                    monitor_resources: bool,  # per workflow
                    run_dir: str) -> Tuple[Callable, Sequence, Dict]:
    """Wrap the Parsl app with a function that will call the monitor function and point it at the correct pid when the task begins.
    """

    @wraps(f)
    def wrapped(*args: List[Any], **kwargs: Dict[str, Any]) -> Any:
        task_id = kwargs.pop('_parsl_monitoring_task_id')
        try_id = kwargs.pop('_parsl_monitoring_try_id')
        terminate_event = Event()
        # Send first message to monitoring router
        send_first_message(try_id,
                           task_id,
                           radio_config,
                           run_id,
                           run_dir)

        if monitor_resources and sleep_dur > 0:
            # create the monitor process and start
            pp = ForkProcess(target=monitor,
                             args=(os.getpid(),
                                   try_id,
                                   task_id,
                                   radio_config,
                                   run_id,
                                   logging_level,
                                   sleep_dur,
                                   run_dir,
                                   terminate_event),
                             daemon=True,
                             name="Monitor-Wrapper-{}".format(task_id))
            pp.start()
            p = pp
            #  TODO: awkwardness because ForkProcess is not directly a constructor
            # and type-checking is expecting p to be optional and cannot
            # narrow down the type of p in this block.

        else:
            p = None

        try:
            return f(*args, **kwargs)
        finally:
            # There's a chance of zombification if the workers are killed by some signals (?)
            if p:
                terminate_event.set()
                p.join(30)  # 30 second delay for this -- this timeout will be hit in the case of an unusually long end-of-loop
                if p.exitcode is None:
                    logger.warn("Event-based termination of monitoring helper took too long. Using process-based termination.")
                    p.terminate()
                    # DANGER: this can corrupt shared queues according to docs.
                    # So, better that the above termination event worked.
                    # This is why this log message is a warning
                    p.join()

            send_last_message(try_id,
                              task_id,
                              radio_config,
                              run_id,
                              run_dir)

    new_kwargs = kwargs.copy()
    new_kwargs['_parsl_monitoring_task_id'] = x_task_id
    new_kwargs['_parsl_monitoring_try_id'] = x_try_id

    return (wrapped, args, new_kwargs)


@wrap_with_logs
def send_first_message(try_id: int,
                       task_id: int,
                       radio_config: RadioConfig,
                       run_id: str, run_dir: str) -> None:
    send_first_last_message(try_id, task_id, radio_config, run_id,
                            run_dir, False)


@wrap_with_logs
def send_last_message(try_id: int,
                      task_id: int,
                      radio_config: RadioConfig,
                      run_id: str, run_dir: str) -> None:
    send_first_last_message(try_id, task_id, radio_config, run_id,
                            run_dir, True)


def send_first_last_message(try_id: int,
                            task_id: int,
                            radio_config: RadioConfig,
                            run_id: str, run_dir: str,
                            is_last: bool) -> None:
    import os
    import platform

    radio = radio_config.create_sender()

    msg = (MessageType.RESOURCE_INFO,
           {'run_id': run_id,
            'try_id': try_id,
            'task_id': task_id,
            'hostname': platform.node(),
            'block_id': os.environ.get('PARSL_WORKER_BLOCK_ID'),
            'first_msg': not is_last,
            'last_msg': is_last,
            'timestamp': datetime.datetime.now()
            })
    radio.send(msg)
    return


@wrap_with_logs
def monitor(pid: int,
            try_id: int,
            task_id: int,
            radio_config: RadioConfig,
            run_id: str,
            logging_level: int,
            sleep_dur: float,
            run_dir: str,
            # removed all defaults because unused and there's no meaningful default for terminate_event.
            # these probably should become named arguments, with a *, and named at invocation.
            terminate_event: Any) -> None:  # cannot be Event because of multiprocessing type weirdness.
    """Monitors the Parsl task's resources by pointing psutil to the task's pid and watching it and its children.

    This process makes calls to logging, but deliberately does not attach
    any log handlers. Previously, there was a handler which logged to a
    file in /tmp, but this was usually not useful or even accessible.
    In some circumstances, it might be useful to hack in a handler so the
    logger calls remain in place.
    """
    import logging
    import platform

    import psutil

    from parsl.utils import setproctitle

    setproctitle("parsl: task resource monitor")

    radio = radio_config.create_sender()

    logging.debug("start of monitor")

    # these values are simple to log. Other information is available in special formats such as memory below.
    simple = ["cpu_num", 'create_time', 'cwd', 'exe', 'memory_percent', 'nice', 'name', 'num_threads', 'pid', 'ppid', 'status', 'username']
    # values that can be summed up to see total resources used by task process and its children
    summable_values = ['memory_percent', 'num_threads']

    pm = psutil.Process(pid)

    children_user_time: Dict[int, float] = {}
    children_system_time: Dict[int, float] = {}
    children_num_ctx_switches_voluntary: Dict[int, float] = {}
    children_num_ctx_switches_involuntary: Dict[int, float] = {}

    def accumulate_and_prepare() -> Dict[str, Any]:
        d = {"psutil_process_" + str(k): v for k, v in pm.as_dict().items() if k in simple}
        d["run_id"] = run_id
        d["task_id"] = task_id
        d["try_id"] = try_id
        d['resource_monitoring_interval'] = sleep_dur
        d['hostname'] = platform.node()
        d['first_msg'] = False
        d['last_msg'] = False
        d['timestamp'] = datetime.datetime.now()

        logging.debug("getting children")
        children = pm.children(recursive=True)
        logging.debug("got children")

        d["psutil_cpu_count"] = psutil.cpu_count()

        # note that this will be the CPU number of the base process, not anything launched by it
        d["psutil_cpu_num"] = pm.cpu_num()

        pctxsw = pm.num_ctx_switches()

        d["psutil_process_num_ctx_switches_voluntary"] = pctxsw.voluntary
        d["psutil_process_num_ctx_switches_involuntary"] = pctxsw.involuntary

        d['psutil_process_memory_virtual'] = pm.memory_info().vms
        d['psutil_process_memory_resident'] = pm.memory_info().rss
        d['psutil_process_time_user'] = pm.cpu_times().user
        d['psutil_process_time_system'] = pm.cpu_times().system
        d['psutil_process_children_count'] = len(children)
        try:
            d['psutil_process_disk_write'] = pm.io_counters().write_chars
            d['psutil_process_disk_read'] = pm.io_counters().read_chars
        except Exception:
            # occasionally pid temp files that hold this information are unvailable to be read so set to zero
            logging.exception("Exception reading IO counters for main process. Recorded IO usage may be incomplete", exc_info=True)
            d['psutil_process_disk_write'] = 0
            d['psutil_process_disk_read'] = 0
        for child in children:
            for k, v in child.as_dict(attrs=summable_values).items():
                d['psutil_process_' + str(k)] += v
            child_user_time = child.cpu_times().user
            child_system_time = child.cpu_times().system
            children_user_time[child.pid] = child_user_time
            children_system_time[child.pid] = child_system_time

            pctxsw = child.num_ctx_switches()
            children_num_ctx_switches_voluntary[child.pid] = pctxsw.voluntary
            children_num_ctx_switches_involuntary[child.pid] = pctxsw.involuntary

            d['psutil_process_memory_virtual'] += child.memory_info().vms
            d['psutil_process_memory_resident'] += child.memory_info().rss
            try:
                d['psutil_process_disk_write'] += child.io_counters().write_chars
                d['psutil_process_disk_read'] += child.io_counters().read_chars
            except Exception:
                # occassionally pid temp files that hold this information are unvailable to be read so add zero
                logging.exception("Exception reading IO counters for child {k}. Recorded IO usage may be incomplete".format(k=k), exc_info=True)
                d['psutil_process_disk_write'] += 0
                d['psutil_process_disk_read'] += 0

        total_children_user_time = 0.0
        for child_pid in children_user_time:
            total_children_user_time += children_user_time[child_pid]

        total_children_system_time = 0.0
        for child_pid in children_system_time:
            total_children_system_time += children_system_time[child_pid]

        total_children_num_ctx_switches_voluntary = 0.0
        for child_pid in children_num_ctx_switches_voluntary:
            total_children_num_ctx_switches_voluntary += children_num_ctx_switches_voluntary[child_pid]

        total_children_num_ctx_switches_involuntary = 0.0
        for child_pid in children_num_ctx_switches_involuntary:
            total_children_num_ctx_switches_involuntary += children_num_ctx_switches_involuntary[child_pid]

        d['psutil_process_time_user'] += total_children_user_time
        d['psutil_process_time_system'] += total_children_system_time
        d['psutil_process_num_ctx_switches_voluntary'] += total_children_num_ctx_switches_voluntary
        d['psutil_process_num_ctx_switches_involuntary'] += total_children_num_ctx_switches_involuntary
        logging.debug("sending message")
        return d

    next_send = time.time()
    accumulate_dur = 5.0  # TODO: make configurable?

    while not terminate_event.is_set() and pm.is_running():
        logging.debug("start of monitoring loop")
        try:
            d = accumulate_and_prepare()
            if time.time() >= next_send:
                logging.debug("Sending intermediate resource message")
                radio.send((MessageType.RESOURCE_INFO, d))
                next_send += sleep_dur
        except Exception:
            logging.exception("Exception getting the resource usage. Not sending usage to Hub", exc_info=True)
        logging.debug("sleeping")

        # wait either until approx next send time, or the accumulation period
        # so the accumulation period will not be completely precise.
        # but before this, the sleep period was also not completely precise.
        # with a minimum floor of 0 to not upset wait

        terminate_event.wait(max(0, min(next_send - time.time(), accumulate_dur)))

    logging.debug("Sending final resource message")
    try:
        d = accumulate_and_prepare()
        radio.send((MessageType.RESOURCE_INFO, d))
    except Exception:
        logging.exception("Exception getting the resource usage. Not sending final usage to Hub", exc_info=True)
    logging.debug("End of monitoring helper")



from enum import Enum


class MessageType(Enum):

    # Reports any task related info such as launch, completion etc.
    TASK_INFO = 0

    # Reports of resource utilization on a per-task basis
    RESOURCE_INFO = 1

    # Top level workflow information
    WORKFLOW_INFO = 2

    # Reports of the resource capacity for each node
    NODE_INFO = 3

    # Reports of the block info
    BLOCK_INFO = 4



from parsl.config import Config
from parsl.executors import HighThroughputExecutor

# from parsl.launchers import SimpleLauncher
from parsl.launchers import SingleNodeLauncher
from parsl.providers import LocalProvider

config = Config(
    executors=[
        HighThroughputExecutor(
            poll_period=1,
            label="htex_local",
            # worker_debug=True,
            cores_per_worker=1,
            encrypted=True,
            provider=LocalProvider(
                init_blocks=1,
                max_blocks=1,
                # tasks_per_node=1,  # For HighThroughputExecutor, this option should in most cases be 1
                launcher=SingleNodeLauncher(),
            ),
        )
    ],
    strategy='none',
)



"""Helpers for cross-platform multiprocessing support.
"""

import logging
import multiprocessing
import multiprocessing.queues
from multiprocessing.context import ForkProcess as ForkProcessType
from multiprocessing.context import SpawnProcess as SpawnProcessType
from typing import Callable

logger = logging.getLogger(__name__)

ForkContext = multiprocessing.get_context("fork")
SpawnContext = multiprocessing.get_context("spawn")

ForkProcess: Callable[..., ForkProcessType] = ForkContext.Process
SpawnProcess: Callable[..., SpawnProcessType] = SpawnContext.Process

SpawnEvent = SpawnContext.Event
SpawnQueue = SpawnContext.Queue


def join_terminate_close_proc(process: SpawnProcessType, *, timeout: int = 30) -> None:
    """Increasingly aggressively terminate a process.

    This function assumes that the process is likely to exit before
    the join timeout, driven by some other means, such as the
    MonitoringHub router_exit_event. If the process does not exit, then
    first terminate() and then kill() will be used to end the process.

    In the case of a very mis-behaving process, this function might take
    up to 3*timeout to exhaust all termination methods and return.
    """
    logger.debug("Joining process")
    process.join(timeout)

    # run a sequence of increasingly aggressive steps to shut down the process.
    if process.is_alive():
        logger.error("Process did not join. Terminating.")
        process.terminate()
        process.join(timeout)
        if process.is_alive():
            logger.error("Process did not join after terminate. Killing.")
            process.kill()
            process.join(timeout)
            # This kill should not be caught by any signal handlers so it is
            # unlikely that this join will timeout. If it does, there isn't
            # anything further to do except log an error in the next if-block.

    if process.is_alive():
        logger.error("Process failed to end")
        # don't call close if the process hasn't ended:
        # process.close() doesn't work on a running process.
    else:
        process.close()



import os

import pytest

import parsl
from parsl.tests.configs.local_radical_mpi import fresh_config as local_config


@parsl.python_app
def some_mpi_func(msg, sleep, comm=None, parsl_resource_specification={}):
    import time
    msg = 'hello %d/%d: %s' % (comm.rank, comm.size, msg)
    time.sleep(sleep)
    print(msg)
    return comm.size


apps = []


@pytest.mark.local
@pytest.mark.radical
@pytest.mark.skipif(
    os.environ.get("RADICAL_CI") != "1",
    reason="Only runs in Radical CI workflow"
)
def test_radical_mpi(n=7):
    # rank size should be > 1 for the
    # radical runtime system to run this function in MPI env
    for i in range(2, n):
        spec = {'ranks': i}
        t = some_mpi_func(msg='mpi.func.%06d' % i, sleep=1, comm=None, parsl_resource_specification=spec)
        apps.append(t)
    assert [len(app.result()) for app in apps] == list(range(2, n))



import logging
import os
import random
from concurrent.futures import Future

import pytest

import parsl


@parsl.python_app
def this_app(inputs=()):
    return inputs[0]


@pytest.mark.local
def test_future_representation(tmpd_cwd):
    import sqlalchemy
    from sqlalchemy import text

    from parsl.tests.configs.htex_local_alternate import fresh_config

    monitoring_db = str(tmpd_cwd / "monitoring.db")
    monitoring_url = "sqlite:///" + monitoring_db

    c = fresh_config()
    c.monitoring.logging_endpoint = monitoring_url
    c.run_dir = str(tmpd_cwd)

    parsl.load(c)

    # we're going to pass this TOKEN into an app via a pre-requisite Future,
    # and then expect to see it appear in the monitoring database.
    TOKEN = random.randint(0, 1000000)

    # make a Future that has no result yet
    # invoke a task that depends on it
    # inspect and insert something about the monitoring recorded value of that Future
    # make the Future complete
    # inspect and insert something about the monitoring recorded value of that Future

    f1 = Future()

    f2 = this_app(inputs=[f1])

    f1.set_result(TOKEN)

    assert f2.result() == TOKEN

    # this cleanup gives a barrier that allows the monitoring code to store
    # everything it has in the database - without this, there's a race
    # condition that the task will not have arrived in the database yet.
    # A different approach for this test might be to poll the DB for a few
    # seconds, with the assumption "data will arrive in the DB within
    # 30 seconds, but probably much sooner".
    parsl.dfk().cleanup()

    engine = sqlalchemy.create_engine(monitoring_url)
    with engine.begin() as connection:
        result = connection.execute(text("SELECT COUNT(*) FROM task"))
        (task_count, ) = result.first()
        assert task_count == 1

        result = connection.execute(text("SELECT task_inputs FROM task"))
        (task_inputs, ) = result.first()
        assert task_inputs == "[" + repr(TOKEN) + "]"



import inspect
from concurrent.futures import Future
from typing import Any, Callable, Dict

import pytest

import parsl
from parsl.executors.base import ParslExecutor

# N is the number of tasks to chain
# With mid-2024 Parsl, N>140 causes Parsl to hang
N = 100

# MAX_STACK is the maximum Python stack depth allowed for either
# task submission to an executor or execution of a task.
# With mid-2024 Parsl, 2-3 stack entries will be used per
# recursively launched parsl task. So this should be smaller than
# 2*N, but big enough to allow regular pytest+parsl stuff to
# happen.
MAX_STACK = 50


def local_config():
    return parsl.Config(executors=[ImmediateExecutor()])


class ImmediateExecutor(ParslExecutor):
    def start(self):
        pass

    def shutdown(self):
        pass

    def submit(self, func: Callable, resource_specification: Dict[str, Any], *args: Any, **kwargs: Any) -> Future:
        stack_depth = len(inspect.stack())
        assert stack_depth < MAX_STACK, "tasks should not be launched deep in the Python stack"
        fut: Future[None] = Future()
        res = func(*args, **kwargs)
        fut.set_result(res)
        return fut


@parsl.python_app
def chain(upstream):
    stack_depth = len(inspect.stack())
    assert stack_depth < MAX_STACK, "chained dependencies should not be launched deep in the Python stack"


@pytest.mark.local
def test_deep_dependency_stack_depth():

    fut = Future()
    here = fut

    for _ in range(N):
        here = chain(here)

    fut.set_result(None)
    here.result()
